"""Constants for pyskyqremote."""
# SOAP/UPnP Constants
SKY_PLAY_URN = "urn:nds-com:serviceId:SkyPlay"
SKYCONTROL = "SkyControl"
SOAP_ACTION = '"urn:schemas-nds-com:service:SkyPlay:2#{0}"'
SOAP_CONTROL_BASE_URL = "http://{0}:49153{1}"
SOAP_DESCRIPTION_BASE_URL = "http://{0}:49153/description{1}.xml"
SOAP_PAYLOAD = """<s:Envelope xmlns:s='http://schemas.xmlsoap.org/soap/envelope/' s:encodingStyle='http://schemas.xmlsoap.org/soap/encoding/'>
    <s:Body>
        <u:{0} xmlns:u="urn:schemas-nds-com:service:SkyPlay:2">
            <InstanceID>0</InstanceID>
        </u:{0}>
    </s:Body>
</s:Envelope>"""
SOAP_RESPONSE = "u:{0}Response"
SOAP_USER_AGENT = "SKYPLUS_skyplus"
UPNP_GET_MEDIA_INFO = "GetMediaInfo"
UPNP_GET_TRANSPORT_INFO = "GetTransportInfo"

# WebSocket Constants
WS_BASE_URL = "ws://{0}:9006/as/{1}"
WS_CURRENT_APPS = "apps/status"

# REST Constants
REST_BASE_URL = "http://{0}:{1}/as/{2}"
REST_CHANNEL_LIST = "services/{0}/{1}"
REST_FAVOURITES = "services/favourites"
REST_RECORDING_DETAILS = "pvr/details/{0}"
REST_RECORDINGS_LIST = "pvr/?limit={0}&offset={1}"
REST_QUOTA_DETAILS = "pvr/storage"
REST_BOOK_RECORDING = "pvr/action/bookrecording?eid={0}"
REST_BOOK_PPVRECORDING = "pvr/action/bookppvrecording?eid={0}&offerref={1}"
REST_BOOK_SERIES_RECORDING = "pvr/action/bookseriesrecording?eid={0}"
REST_SERIES_LINK = "pvr/action/serieslink?pvrid={0}"
REST_SERIES_UNLINK = "pvr/action/seriesunlink?pvrid={0}"
REST_RECORDING_KEEP = "pvr/action/keep?pvrid={0}"
REST_RECORDING_UNKEEP = "pvr/action/unkeep?pvrid={0}"
REST_RECORDING_LOCK = "pvr/action/lock?pvrid={0}"
REST_RECORDING_UNLOCK = "pvr/action/unlock?pvrid={0}"
REST_RECORDING_DELETE = "pvr/action/delete?pvrid={0}"
REST_RECORDING_UNDELETE = "pvr/action/undelete?pvrid={0}"
REST_RECORDING_ERASE = "pvr/action/erase?pvrid={0}"
REST_RECORDING_ERASE_ALL = "pvr"
REST_RECORDING_SET_LAST_PLAYED_POSITION = "pvr/action/setlastplayedposition?pos={0}&pvrid={1}"
REST_PATH_SYSTEMINFO = "system/information"
REST_PATH_DEVICEINFO = "system/deviceinformation"
REST_PATH_APPS = "apps"
REST_GET = "get"
REST_POST = "post"
REST_DELETE = "delete"

# Sky specific constants
CURRENT_URI = "CurrentURI"
CURRENT_TRANSPORT_STATE = "CurrentTransportState"
APP_STATUS_VISIBLE = "VISIBLE"

PVR = "pvr"
XSI = "xsi"

SKY_STATE_NOMEDIA = "NO_MEDIA_PRESENT"
SKY_STATE_OFF = "POWERED OFF"
SKY_STATE_ON = "ON"
SKY_STATE_PLAYING = "PLAYING"
SKY_STATE_PAUSED = "PAUSED_PLAYBACK"
SKY_STATE_STANDBY = "STANDBY"
SKY_STATE_STOPPED = "STOPPED"
SKY_STATE_TRANSITIONING = "TRANSITIONING"
SKY_STATUS_LIVE = "LIVE"

APP_EPG = "com.bskyb.epgui"

COMMANDS = {
    "power": 0,
    "select": 1,
    "backup": 2,
    "dismiss": 2,
    "channelup": 6,
    "channeldown": 7,
    "interactive": 8,
    "sidebar": 8,
    "help": 9,
    "services": 10,
    "search": 10,
    "tvguide": 11,
    "home": 11,
    "i": 14,
    "text": 15,
    "up": 16,
    "down": 17,
    "left": 18,
    "right": 19,
    "red": 32,
    "green": 33,
    "yellow": 34,
    "blue": 35,
    "0": 48,
    "1": 49,
    "2": 50,
    "3": 51,
    "4": 52,
    "5": 53,
    "6": 54,
    "7": 55,
    "8": 56,
    "9": 57,
    "play": 64,
    "pause": 65,
    "stop": 66,
    "record": 67,
    "fastforward": 69,
    "rewind": 71,
    "boxoffice": 240,
    "sky": 241,
}

KNOWN_COUNTRIES = {
    "DEU": "DEU",
    "GBR": "GBR",
    "IRL": "GBR",
    "ITA": "ITA",
}

# Random set of other constants
EPG_ERROR_PAST_END = "past end of epg"
EPG_ERROR_NO_DATA = "no epg data found"

RESPONSE_OK = 200
CONNECT_TIMEOUT = 1000
HTTP_TIMEOUT = 6
SOAP_TIMEOUT = 2

AUDIO = "audio"
VIDEO = "video"

ALLRECORDINGS = "all"

REST_GET = "get"
REST_POST = "post"
REST_DELETE = "DELETE"



"""UK specific code."""
import logging
from datetime import datetime

import requests

from ..classes.programme import Programme
from ..const import RESPONSE_OK, SKY_STATUS_LIVE
from .const_gb import CHANNEL_IMAGE_URL, LIVE_IMAGE_URL, PVR_IMAGE_URL, SCHEDULE_URL, TERRITORY

_LOGGER = logging.getLogger(__name__)


class SkyQCountry:
    """UK specific SkyQ."""

    def __init__(self):
        """Initialise UK remote."""
        self.pvr_image_url = PVR_IMAGE_URL

    def getEpgData(self, sid, channelno, channelName, epgDate):
        """Get EPG data for UK."""
        return self._getData(sid, channelno, channelName, epgDate)

    def buildChannelImageUrl(self, sid, channelname):
        """Build the channel image URL."""
        chid = "".join(e for e in channelname.casefold() if e.isalnum())
        return CHANNEL_IMAGE_URL.format(sid, chid)

    def _getData(self, sid, channelno, channelName, epgDate):
        epgDateStr = epgDate.strftime("%Y%m%d")

        epgUrl = SCHEDULE_URL.format(sid, epgDateStr)
        headers = {"x-skyott-territory": TERRITORY, "x-skyott-provider": "SKY", "x-skyott-proposition": "SKYQ"}
        programmes = set()

        resp = requests.get(epgUrl, headers=headers)
        epgData = resp.json()["schedule"] if resp.status_code == RESPONSE_OK else None
        if epgData is None:
            return programmes

        if len(epgData) == 0:
            return programmes

        for p in epgData[0]["events"]:
            starttime = datetime.utcfromtimestamp(p["st"])
            endtime = datetime.utcfromtimestamp(p["st"] + p["d"])
            title = p["t"]
            season = None
            if "seasonnumber" in p and p["seasonnumber"] > 0:
                season = p["seasonnumber"]
            episode = None
            if "episodenumber" in p and p["episodenumber"] > 0:
                episode = p["episodenumber"]
            programmeuuid = None
            imageUrl = None
            if "programmeuuid" in p:
                programmeuuid = str(p["programmeuuid"])
                imageUrl = LIVE_IMAGE_URL.format(programmeuuid)

            eid = p["eid"]
            programme = Programme(
                programmeuuid,
                starttime,
                endtime,
                title,
                season,
                episode,
                imageUrl,
                channelName,
                SKY_STATUS_LIVE,
                "n/a",
                eid,
            )
            programmes.add(programme)

        return programmes



"""Constants for pyskyqremote- IT."""

SCHEDULE_URL = "https://apid.sky.it/gtv/v1/events?from={1}&to={2}&pageSize=50&pageNum=0&env=DTH&channels={0}"
LIVE_IMAGE_URL = "https://ethaneurope.it.imageservice.sky.com/pd-image/{0}/16-9"
PVR_IMAGE_URL = "https://ethaneurope.it.imageservice.sky.com/pd-image/{0}/16-9"
CHANNEL_URL = "https://apid.sky.it/gtv/v1/channels?env=DTH"
CHANNEL_IMAGE_URL = (
    "https://ethaneurope.it.imageservice.sky.com/pd-logo/skychb_{0}{1}/600/600"
)



"""Information for channellist and channel."""

import json
from dataclasses import dataclass, field
from operator import attrgetter

from ..const import AUDIO, REST_CHANNEL_LIST, VIDEO
from ..const_test import TEST_CHANNEL_LIST


class ChannelInformation:
    """Channel information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the channel information class."""
        self._remoteConfig = remoteConfig
        self._deviceAccess = remoteConfig.deviceAccess
        self._test_channel = remoteConfig.test_channel
        self._channels = []
        self._bouquet = remoteConfig.deviceInfo.bouquet
        self._subbouquet = remoteConfig.deviceInfo.subbouquet

    def getChannelList(self):
        """Get Channel list for Sky Q box."""
        self._channels = self._getChannels()
        if not self._channels:
            return None

        channelitems = set()

        for c in self._channels:
            channel = Channel(c["c"], c["t"], c["sid"], None, sf=c["sf"])
            channelitems.add(channel)

        channelnosorted = sorted(channelitems, key=attrgetter("channelnoint"))

        return ChannelList(sorted(channelnosorted, key=attrgetter("channeltype"), reverse=True))

    def getChannelInfo(self, channelNo):
        """Retrieve channel information for specified channelNo."""
        if not channelNo.isnumeric():
            return None

        if not self._channels:
            self._channels = self._getChannels()

        try:
            channel = next(c for c in self._channels if c["c"] == channelNo)
        except StopIteration:
            return None

        channelno = channel["c"]
        channelname = channel["t"]
        channelsid = channel["sid"]
        channelImageUrl = self._remoteConfig.remoteCountry.buildChannelImageUrl(channelsid, channelname)
        sf = channel["sf"]
        return Channel(channelno, channelname, channelsid, channelImageUrl, sf=sf)

    def getChannelNode(
        self,
        sid,
    ):
        """Retrieve the channel node for the given sid."""
        channelNode = self._getNodeFromChannels(sid)

        if not channelNode:
            # Load the channel list for the first time.
            # It's also possible the channels may have changed since last HA restart, so reload them
            self._channels = self._getChannels()
            channelNode = self._getNodeFromChannels(sid)
        if not channelNode:
            return None

        channel = channelNode["t"]
        channelno = channelNode["c"]
        return {"channel": channel, "channelno": channelno}

    def _getChannels(self):
        """Get the list of channels from the Sky Q box."""
        # This is here because otherwise I can never validate code for a foreign device
        if self._test_channel:
            return TEST_CHANNEL_LIST

        channels = self._deviceAccess.retrieveInformation(REST_CHANNEL_LIST.format(self._bouquet, self._subbouquet))
        if channels and "services" in channels:
            return channels["services"]

        return []

    def _getNodeFromChannels(self, sid):
        return next((s for s in self._channels if s["sid"] == str(sid)), None)


@dataclass
class ChannelList:
    """SkyQ Channel List Class."""

    channels: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing the Channel list."""
        return json.dumps(self, cls=_ChannelListJSONEncoder)


def ChannelListDecoder(obj):
    """Decode the channel list object from json."""
    channellist = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in channellist and channellist["__type__"] == "__channellist__":
        return ChannelList(channels=channellist["channels"], **channellist["attributes"])
    return channellist


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "__type__" in obj and obj["__type__"] == "__channel__":
        obj = Channel(**obj["attributes"])
    return obj


class _ChannelListJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ChannelList):
            type_ = "__channellist__"
            channels = obj.channels
            attributes = {k: v for k, v in vars(obj).items() if k not in {"channels"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "channels": channels,
            }

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, Channel):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__channel__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, obj)  # pragma: no cover


@dataclass(order=True)
class Channel:
    """SkyQ Channel Class."""

    channelno: str = field(
        init=True,
        repr=True,
        compare=True,
    )
    channelname: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelsid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelimageurl: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channeltype: str = None
    channelnoint: int = None
    sf: str = None

    def __post_init__(self):
        """Post process the channel setup."""
        self.channeltype = AUDIO if self.sf == "au" else VIDEO
        self.channelnoint = int(self.channelno)

    def __hash__(self):
        """Calculate the hash of this object."""
        typesort = "100" if self.channeltype == VIDEO else "20"
        return hash(typesort + self.channelno)

    def as_json(self) -> str:
        """Return a JSON string representing this Channel."""
        return json.dumps(self, cls=_ChannelJSONEncoder)


def ChannelDecoder(obj):
    """Decode channel object from json."""
    channel = json.loads(obj)
    if "__type__" in channel and channel["__type__"] == "__channel__":
        return Channel(**channel["attributes"])
    return channel


class _ChannelJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Channel):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__channel__",
                "attributes": attributes,
            }



"""Favourites methods for the Sky Q box."""
import json
from dataclasses import dataclass, field
from operator import attrgetter

from ..const import REST_FAVOURITES


class FavouriteInformation:
    """Sky Q favourites information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the favourites information class."""
        self._deviceAccess = remoteConfig.deviceAccess

    def getFavouriteList(self, channellist):
        """Retrieve the list of favourites."""
        self._channellist = channellist
        favourites = self._deviceAccess.retrieveInformation(REST_FAVOURITES)
        if not favourites or "favourites" not in favourites:
            return []

        favitems = set()

        for f in favourites["favourites"]:
            favsid = f["sid"]
            channel = self._getFavChannel(favsid)
            channelno = channel.channelno if channel else None
            channelname = channel.channelname if channel else None
            favourite = Favourite(f["lcn"], channelno, channelname, favsid)
            favitems.add(favourite)

        favouritesorted = sorted(favitems, key=attrgetter("lcn"))
        self._favouritelist = FavouriteList(favouritesorted)

        return self._favouritelist

    def _getFavChannel(self, sid):
        try:
            return next(c for c in self._channellist.channels if c.channelsid == sid)
        except StopIteration:
            return None


@dataclass
class FavouriteList:
    """SkyQ Favourite List Class."""

    favourites: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this Favourite."""
        return json.dumps(self, cls=_FavouriteListJSONEncoder)


def FavouriteListDecoder(obj):
    """Decode favourite object from json."""
    favouritelist = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in favouritelist and favouritelist["__type__"] == "__favouritelist__":
        return FavouriteList(favourites=favouritelist["favourites"], **favouritelist["attributes"])
    return favouritelist


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "__type__" in obj and obj["__type__"] == "__favourite__":
        obj = Favourite(**obj["attributes"])
    return obj


class _FavouriteListJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, FavouriteList):
            type_ = "__favouritelist__"
            favourites = obj.favourites
            attributes = {k: v for k, v in vars(obj).items() if k not in {"favourites"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "favourites": favourites,
            }

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, Favourite):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__favourite__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, obj)  # pragma: no cover


@dataclass(order=True)
class Favourite:
    """SkyQ favourite Class."""

    lcn: int = field(
        init=True,
        repr=True,
        compare=True,
    )
    channelno: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelname: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    sid: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def __hash__(self):
        """Calculate the hash of this object."""
        return hash(self.lcn)

    def as_json(self) -> str:
        """Return a JSON string representing this favourite."""
        return json.dumps(self, cls=_favouriteJSONEncoder)


def favouriteDecoder(obj):
    """Decode favourite object from json."""
    favourite = json.loads(obj)
    if "__type__" in favourite and favourite["__type__"] == "__favourite__":
        return favourite(**favourite["attributes"])
    return favourite


class _favouriteJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Favourite):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__favourite__",
                "attributes": attributes,
            }



"""Constants for pyskyqremote- GB."""
TERRITORY = "GB"

SCHEDULE_URL = "http://atlantis.epgsky.com/as/schedule/{1}/{0}"
LIVE_IMAGE_URL = (
    "https://imageservice.sky.com/pd-image/{0}/16-9/456?territory=" + TERRITORY + "&provider=SKY&proposition=SKYQ"
)
PVR_IMAGE_URL = LIVE_IMAGE_URL
CHANNEL_IMAGE_URL = (
    "https://imageservice.sky.com/logo/skychb_{0}{1}/600/600?territory=" + TERRITORY + "&provider=SKY&proposition=SKYQ"
)



"""Python module for accessing SkyQ box and EPG, and sending commands."""
import importlib
import logging
from datetime import datetime

import pycountry

from .classes.app import AppInformation
from .classes.channel import ChannelInformation
from .classes.channelepg import ChannelEPGInformation
from .classes.device import DeviceInformation
from .classes.deviceaccess import DeviceAccess
from .classes.favourite import FavouriteInformation
from .classes.media import MediaInformation
from .classes.programme import Programme
from .classes.recordings import RecordingsInformation
from .const import (
    ALLRECORDINGS,
    COMMANDS,
    CURRENT_TRANSPORT_STATE,
    EPG_ERROR_NO_DATA,
    EPG_ERROR_PAST_END,
    SKY_STATE_NOMEDIA,
    SKY_STATE_OFF,
    SKY_STATE_ON,
    SKY_STATE_PAUSED,
    SKY_STATE_PLAYING,
    SKY_STATE_STANDBY,
    SKY_STATE_STOPPED,
    SKY_STATE_TRANSITIONING,
)

_LOGGER = logging.getLogger(__name__)


class SkyQRemote:
    """SkyQRemote is the instantiation of the SKYQ remote ccontrol."""

    commands = COMMANDS

    def __init__(self, host, epgCacheLen=20, port=49160, jsonPort=9006):
        """Stand up a new SkyQ box."""
        self.deviceSetup = False
        self.gateway = False
        self._host = host
        self._remoteCountry = None
        self._overrideCountry = None
        self._channel = None
        self._programme = None
        self._recordedProgramme = None
        self._lastProgrammeEpg = None
        self._lastPvrId = None
        self._error = False
        self._channellist = None

        self._appInformation = None
        self._channelInformation = None
        self._deviceInformation = None
        self._channelEPGInformation = None
        self._favouriteInformation = None
        self._mediaInformation = None
        self._recordingsInformation = None

        self._remoteConfig = _RemoteConfig(host, port, jsonPort, epgCacheLen)

        deviceInfo = self.getDeviceInformation()
        if not deviceInfo:
            return None

        self._setupDevice()

    def powerStatus(self) -> str:
        """Get the power status of the Sky Q box."""
        if not self._remoteCountry:
            self._setupRemote()

        if self._remoteConfig.soapControlURL is None:
            return SKY_STATE_OFF

        systemInfo = self._deviceInformation.getSystemInformation()

        if systemInfo is None:
            return SKY_STATE_OFF
        if "activeStandby" in systemInfo and systemInfo["activeStandby"] is True:
            return SKY_STATE_STANDBY

        return SKY_STATE_ON

    def getCurrentState(self):
        """Get current state of the SkyQ box."""
        if not self._remoteCountry:
            self._setupRemote()

        if self._remoteConfig.soapControlURL is None:
            return SKY_STATE_OFF

        response = self._deviceInformation.getTransportInformation()
        if response is not None:
            state = response[CURRENT_TRANSPORT_STATE]
            if state in [SKY_STATE_NOMEDIA, SKY_STATE_STOPPED]:
                return SKY_STATE_STANDBY
            if state in [SKY_STATE_PLAYING, SKY_STATE_TRANSITIONING]:
                return SKY_STATE_PLAYING
            if state == SKY_STATE_PAUSED:
                return SKY_STATE_PAUSED
        return SKY_STATE_OFF

    def getActiveApplication(self):
        """Get the active application on Sky Q box."""
        if not self._appInformation:
            self._appInformation = AppInformation(self._remoteConfig)

        return self._appInformation.getActiveApplication()

    def getCurrentMedia(self):
        """Get the currently playing media on the SkyQ box."""
        if not self._mediaInformation:
            self._mediaInformation = MediaInformation(self._remoteConfig)

        return self._mediaInformation.getCurrentMedia()

    def getEpgData(self, sid, epgDate, days=2):
        """Get EPG data for the specified channel/date."""
        if not self._channelEPGInformation:
            self._channeEPGInformation = ChannelEPGInformation(self._remoteConfig)

        return self._channeEPGInformation.getEpgData(sid, epgDate, days)

    def getProgrammeFromEpg(self, sid, epgDate, queryDate):
        """Get programme from EPG for specfied time and channel."""
        sidint = 0
        try:
            sidint = int(sid)
        except ValueError:
            if not self._error:
                self._error = True
                _LOGGER.info(
                    f"I0010 - Programme data not found for host: {self._host}/{self._overrideCountry} sid: {sid} : {epgDate}"
                )
                return EPG_ERROR_NO_DATA

            self._error = False

        programmeEpg = f"{str(sidint)} {epgDate.strftime('%Y%m%d')}"
        if self._lastProgrammeEpg == programmeEpg and queryDate < self._programme.endtime:
            return self._programme

        epgData = self.getEpgData(sidint, epgDate)

        if len(epgData.programmes) == 0:
            if not self._error:
                self._error = True
                _LOGGER.info(
                    f"I0020 - Programme data not found for host: {self._host}/{self._overrideCountry} sid: {sid} : {epgDate}"
                )
            return EPG_ERROR_NO_DATA

        self._error = False

        try:
            programme = next(p for p in epgData.programmes if p.starttime <= queryDate and p.endtime >= queryDate)

            self._programme = programme
            self._lastProgrammeEpg = programmeEpg
            return programme

        except StopIteration:
            return EPG_ERROR_PAST_END

    def getCurrentLiveTVProgramme(self, sid):
        """Get current live programme on the specified channel."""
        try:
            queryDate = datetime.utcnow()
            programme = self.getProgrammeFromEpg(sid, queryDate, queryDate)
            if not isinstance(programme, Programme):
                return None

            return programme
        except Exception as err:
            _LOGGER.exception(f"X0010 - Error occurred: {self._host} : {sid} : {err}")
            return None

    def getRecordings(self, status=ALLRECORDINGS, limit=1000, offset=0):
        """Get the list of available Recordings."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.getRecordings(status, limit, offset)

    def getRecording(self, pvrId):
        """Get the recording details."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        if self._lastPvrId == pvrId:
            return self._recordedProgramme
        self._lastPvrId = pvrId

        self._recordedProgramme = self._recordingsInformation.getRecording(pvrId)
        return self._recordedProgramme

    def getQuota(self):
        """Retrieve the available storage quota."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.getQuota()

    def bookRecording(self, eid, series=False):
        """Book recording for specified item."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.bookRecording(eid, series)

    def bookPPVRecording(self, eid, offerref):
        """Book recording for specified item."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.bookPPVRecording(eid, offerref)

    def seriesLink(self, pvrid, On=True):
        """Book recording for specified item."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.seriesLink(pvrid, On)

    def recordingKeep(self, pvrid, On=True):
        """Keep the recording."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingKeep(pvrid, On)

    def recordingLock(self, pvrid, On=True):
        """Lock the recording."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingLock(pvrid, On)

    def recordingDelete(self, pvrid, On=True):
        """Delete the recording."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingDelete(pvrid, On)

    def recordingErase(self, pvrid):
        """Delete the recording."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingErase(pvrid)

    def recordingEraseAll(self):
        """Delete the reording."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingEraseAll()

    def recordingSetLastPlayedPosition(self, pvrid, pos):
        """Set the last played position for specified item."""
        if not self._recordingsInformation:
            self._recordingsInformation = RecordingsInformation(self._remoteConfig)

        return self._recordingsInformation.recordingSetLastPlayedPosition(pvrid, pos)

    def getDeviceInformation(self):
        """Get the device information from the SkyQ box."""
        if not self._deviceInformation:
            self._deviceInformation = DeviceInformation(self._remoteConfig)

        deviceInfo = self._deviceInformation.getDeviceInformation(self._overrideCountry)
        self._remoteConfig.deviceInfo = deviceInfo
        return deviceInfo

    def getChannelList(self):
        """Get Channel list for Sky Q box."""
        if not self._channelInformation:
            self._channelInformation = ChannelInformation(self._remoteConfig)

        self._channellist = self._channelInformation.getChannelList()
        return self._channellist

    def getChannelInfo(self, channelNo):
        """Retrieve channel information for specified channelNo."""
        if not self._channelInformation:
            self._channelInformation = ChannelInformation(self._remoteConfig)

        return self._channelInformation.getChannelInfo(channelNo)

    def getFavouriteList(self):
        """Retrieve the list of favourites."""
        if not self._favouriteInformation:
            self._favouriteInformation = FavouriteInformation(self._remoteConfig)

        if not self._channellist:
            self.getChannelList()
        return self._favouriteInformation.getFavouriteList(self._channellist)

    def press(self, sequence):
        """Issue the specified sequence of commands to SkyQ box."""
        self._remoteConfig.deviceAccess.press(sequence)

    def setOverrides(self, overrideCountry=None, test_channel=None, jsonPort=None, port=None):
        """Override various items."""
        if overrideCountry:
            self._overrideCountry = overrideCountry
        if test_channel:
            self._remoteConfig.test_channel = test_channel
        if jsonPort:
            self._remoteConfig.jsonPort = jsonPort
        if port:
            self._remoteConfig.port = port

    def _setupRemote(self):
        deviceInfo = self.getDeviceInformation()
        if not deviceInfo:
            return

        if not self.deviceSetup:
            self._setupDevice()

        if not self._remoteCountry and self.deviceSetup:
            SkyQCountry = self._importCountry(self._remoteConfig.deviceInfo.epgCountryCode)
            self._remoteCountry = SkyQCountry()
            self._remoteConfig.remoteCountry = self._remoteCountry

    def _setupDevice(self):

        self._remoteConfig.soapControlURL = self._deviceInformation.getSoapControlURL()

        self.deviceSetup = True
        self.gateway = self._remoteConfig.deviceInfo.gateway

    def _importCountry(self, countryCode):
        """Work out the country for the Country Code."""
        try:
            country = pycountry.countries.get(alpha_3=countryCode).alpha_2.casefold()
            SkyQCountry = importlib.import_module("pyskyqremote.country.remote_" + country).SkyQCountry

        except (AttributeError, ModuleNotFoundError) as err:
            _LOGGER.warning(f"W0010 - Invalid country, defaulting to GBR: {self._host} : {countryCode} : {err}")

            from pyskyqremote.country.remote_gb import SkyQCountry

        return SkyQCountry


class _RemoteConfig:
    deviceAccess = None
    host = 0
    port = 0
    jsonPort = 0
    epgCacheLen = 0
    remoteCountry = ""
    test_channel = 0
    soapControlURL = ""
    epgCacheLen = 0
    deviceInfo = None

    def __init__(
        self,
        host,
        port,
        jsonPort,
        epgCacheLen,
        deviceAccess=None,
        remoteCountry=None,
        test_channel=None,
        soapControlURL=None,
        deviceInfo=None,
    ):
        self.host = host
        self.port = port
        self.jsonPort = jsonPort
        self.deviceAccess = deviceAccess
        self.remoteCountry = remoteCountry
        self.test_channel = test_channel
        self.soapControlURL = soapControlURL
        self.epgCacheLen = epgCacheLen
        self.deviceAccess = DeviceAccess(host, jsonPort, port)
        self.deviceInfo = deviceInfo



"""Sky Q app methods and data class."""

import json
from dataclasses import dataclass, field

from ..const import APP_EPG, APP_STATUS_VISIBLE, REST_PATH_APPS, WS_CURRENT_APPS


class AppInformation:
    """Sky Q app information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the app information class."""
        self._deviceAccess = remoteConfig.deviceAccess
        self._currentApp = APP_EPG
        self._apps = {}

    def getActiveApplication(self):
        """Get the active application on Sky Q box."""
        try:
            apps = self._deviceAccess.callSkyWebSocket(WS_CURRENT_APPS)
            if apps:
                self._currentApp = next(a for a in apps["apps"] if a["status"] == APP_STATUS_VISIBLE)["appId"]

            return App(self._currentApp, self._get_app_title(self._currentApp))
        except Exception:
            return App(self._currentApp, self._get_app_title(self._currentApp))

    def _get_app_title(self, appId):
        if len(self._apps) == 0:
            apps = self._deviceAccess.retrieveInformation(REST_PATH_APPS)
            if not apps:
                return None
            for a in apps["apps"]:
                self._apps[a["appId"]] = a["title"]

        return self._apps[appId] if appId in self._apps else None


@dataclass
class App:
    """SkyQ App Class."""

    appId: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    title: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this app info."""
        return json.dumps(self, cls=_AppJSONEncoder)


def AppDecoder(obj):
    """Decode programme object from json."""
    app = json.loads(obj)
    if "__type__" in app and app["__type__"] == "__app__":
        return App(**app["attributes"])
    return app


class _AppJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, App):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__app__",
                "attributes": attributes,
            }



"""Italy specific code."""
import logging
from datetime import datetime, timedelta

import requests

from ..classes.programme import Programme
from ..const import RESPONSE_OK, SKY_STATUS_LIVE
from .const_it import CHANNEL_IMAGE_URL, CHANNEL_URL, LIVE_IMAGE_URL, PVR_IMAGE_URL, SCHEDULE_URL

_LOGGER = logging.getLogger(__name__)


class SkyQCountry:
    """Italy specific SkyQ."""

    def __init__(self):
        """Initialise Italy remote."""
        self.pvr_image_url = PVR_IMAGE_URL
        self._channellist = None

        self._getChannels()

    def getEpgData(self, sid, channelno, channelName, epgDate):
        """Get EPG data for Italy."""
        epgPrev = epgDate - timedelta(days=1)
        queryDateFrom = epgPrev.strftime("%Y-%m-%dT22:00:00Z")
        queryDateTo = epgDate.strftime("%Y-%m-%dT23:59:59Z")
        epgData = self._getData(sid, channelno, channelName, queryDateFrom, queryDateTo)

        midnight = datetime.combine(epgDate.date(), datetime.min.time())

        return [p for p in epgData if p.endtime >= midnight]

    def buildChannelImageUrl(self, sid, channelname):
        """Build the channel image URL."""
        chid = "".join(e for e in channelname.casefold() if e.isalnum())
        return CHANNEL_IMAGE_URL.format(sid, chid)

    def _getData(self, sid, channelno, channelName, queryDateFrom, queryDateTo):
        cid = None
        for channel in self._channellist:
            if str(channel["number"]) == str(channelno):
                cid = channel["id"]

        epgUrl = SCHEDULE_URL.format(cid, queryDateFrom, queryDateTo)
        programmes = set()

        try:
            resp = requests.get(epgUrl)
        except requests.exceptions.ConnectionError:
            return programmes

        epgData = resp.json()["events"] if resp.status_code == RESPONSE_OK else None
        if epgData is None:
            return programmes

        if len(epgData) == 0:
            return programmes

        epgDataLen = len(epgData) - 1
        for index, p in enumerate(epgData):
            starttime = datetime.strptime(p["starttime"], "%Y-%m-%dT%H:%M:%SZ")
            if index < epgDataLen:
                endtimeStr = epgData[index + 1]["starttime"]
            else:
                endtimeStr = p["endtime"]
            endtime = datetime.strptime(endtimeStr, "%Y-%m-%dT%H:%M:%SZ")
            title = p["eventTitle"]
            season = None
            if "seasonNumber" in p["content"] and p["content"]["seasonNumber"] > 0:
                season = p["content"]["seasonNumber"]
            episode = None
            if "episodeNumber" in p["content"] and p["content"]["episodeNumber"] > 0:
                episode = p["content"]["episodeNumber"]
            programmeuuid = None
            imageUrl = None
            if "uuid" in p["content"]:
                programmeuuid = str(p["content"]["uuid"])
                imageUrl = LIVE_IMAGE_URL.format(programmeuuid)

            programme = Programme(
                programmeuuid,
                starttime,
                endtime,
                title,
                season,
                episode,
                imageUrl,
                channelName,
                SKY_STATUS_LIVE,
            )
            programmes.add(programme)

        return programmes

    def _getChannels(self):
        resp = requests.get(CHANNEL_URL)
        if resp.status_code == RESPONSE_OK:
            self._channellist = resp.json()["channels"]



"""Structure of a standard EPG programme."""

import json
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime, timedelta

from .channel import ChannelInformation
from .programme import Programme


class ChannelEPGInformation:
    """Sky Q Channel EPG information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the channel epg information class."""
        self._remoteConfig = remoteConfig
        self._deviceAccess = remoteConfig.deviceAccess
        self._remoteCountry = remoteConfig.remoteCountry
        self._test_channel = remoteConfig.test_channel
        self._epgCacheLen = remoteConfig.epgCacheLen
        self._epgCache = OrderedDict()
        self._channel = None
        self._channelInformation = None

    def getEpgData(self, sid, epgDate, days=2):
        """Get EPG data for the specified channel/date."""
        epg = f"{str(sid)} {'{:0>2d}'.format(days)} {epgDate.strftime('%Y%m%d')}"

        if sid in self._epgCache and self._epgCache[sid]["epg"] == epg:
            return self._epgCache[sid]["channel"]

        channelNo = None
        channelName = None
        channelImageUrl = None
        programmes = set()

        channelNode = self._getChannelNode(sid)
        if channelNode:
            channelNo = channelNode["channelno"]
            channelName = channelNode["channel"]
            channelImageUrl = self._remoteCountry.buildChannelImageUrl(sid, channelName)

            for n in range(days):
                programmesData = self._remoteCountry.getEpgData(
                    sid, channelNo, channelName, epgDate + timedelta(days=n)
                )
                if len(programmesData) > 0:
                    programmes = programmes.union(programmesData)
                else:
                    break

        self._channel = ChannelEPG(sid, channelNo, channelName, channelImageUrl, sorted(programmes))
        self._epgCache[sid] = {
            "epg": epg,
            "channel": self._channel,
            "updatetime": datetime.utcnow(),
        }
        self._epgCache = OrderedDict(sorted(self._epgCache.items(), key=lambda x: x[1]["updatetime"], reverse=True))
        while len(self._epgCache) > self._epgCacheLen:
            self._epgCache.popitem(last=True)

        return self._channel

    def _getChannelNode(self, sid):
        if not self._channelInformation:
            self._channelInformation = ChannelInformation(self._remoteConfig)

        return self._channelInformation.getChannelNode(sid)


@dataclass
class ChannelEPG:
    """SkyQ Channel EPG Class."""

    sid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelno: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelname: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelImageUrl: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    programmes: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this Channel."""
        return json.dumps(self, cls=_ChannelEPGJSONEncoder)


def ChannelEPGDecoder(obj):
    """Decode channel object from json."""
    channelepg = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in channelepg and channelepg["__type__"] == "__channelepg__":
        return ChannelEPG(programmes=channelepg["programmes"], **channelepg["attributes"])
    return channelepg


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "starttime" in obj:
        obj["starttime"] = datetime.strptime(obj["starttime"], "%Y-%m-%dT%H:%M:%SZ")
    if "endtime" in obj:
        obj["endtime"] = datetime.strptime(obj["endtime"], "%Y-%m-%dT%H:%M:%SZ")
    if "__type__" in obj and obj["__type__"] == "__programme__":
        obj = Programme(**obj["attributes"])
    return obj


class _ChannelEPGJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ChannelEPG):
            type_ = "__channelepg__"
            programmes = obj.programmes
            attributes = {k: v for k, v in vars(obj).items() if k not in {"programmes"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "programmes": programmes,
            }

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, Programme):
            attributes = {}
            for k, v in vars(obj).items():
                if isinstance(v, datetime):
                    v = v.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = v
            return {
                "__type__": "__programme__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, obj)  # pragma: no cover



"""Structure of a media information."""

import json
from dataclasses import dataclass, field
from datetime import datetime

from ..const import CURRENT_URI, PVR, UPNP_GET_MEDIA_INFO, XSI
from .channel import ChannelInformation


class MediaInformation:
    """Sky Q media information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the media information class."""
        self._remoteConfig = remoteConfig
        self._deviceAccess = remoteConfig.deviceAccess
        self._soapControlURL = remoteConfig.soapControlURL
        self._remoteCountry = remoteConfig.remoteCountry
        self._test_channel = remoteConfig.test_channel
        self._channelInformation = None

    def getCurrentMedia(self):
        """Get the currently playing media on the SkyQ box."""
        channel = None
        channelno = None
        imageUrl = None
        sid = None
        pvrId = None
        live = False

        response = self._deviceAccess.callSkySOAPService(self._soapControlURL, UPNP_GET_MEDIA_INFO)
        if response is None:
            return None

        currentURI = response[CURRENT_URI]
        if currentURI is None:
            return None

        if XSI in currentURI:
            sid = self._test_channel or int(currentURI[6:], 16)
            live = True
            channelNode = self._getChannelNode(sid)
            if channelNode:
                channel = channelNode["channel"]
                channelno = channelNode["channelno"]
                imageUrl = self._remoteCountry.buildChannelImageUrl(sid, channel)
        elif PVR in currentURI:
            # Recorded content
            pvrId = "P" + currentURI[11:]
            live = False

        return Media(channel, channelno, imageUrl, sid, pvrId, live)

    def _getChannelNode(self, sid):
        if not self._channelInformation:
            self._channelInformation = ChannelInformation(self._remoteConfig)

        return self._channelInformation.getChannelNode(sid)


@dataclass
class Media:
    """SkyQ Programme Class."""

    channel: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelno: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    imageUrl: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    sid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    pvrId: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    live: bool = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this media info."""
        return json.dumps(self, cls=_MediaJSONEncoder)


def MediaDecoder(obj):
    """Decode programme object from json."""
    media = json.loads(obj)
    if "__type__" in media and media["__type__"] == "__media__":
        return Media(**media["attributes"])
    return media


class _MediaJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Media):
            attributes = {}
            for k, v in vars(obj).items():
                if isinstance(v, datetime):
                    v = v.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = v
            return {
                "__type__": "__media__",
                "attributes": attributes,
            }



"""SKY Q Remote Utilities."""
import json
import logging
import math
import socket
import time
from http import HTTPStatus

import requests
import websocket
import xmltodict

from ..const import (
    COMMANDS,
    CONNECT_TIMEOUT,
    HTTP_TIMEOUT,
    REST_BASE_URL,
    REST_DELETE,
    REST_GET,
    REST_POST,
    SKY_PLAY_URN,
    SKYCONTROL,
    SOAP_ACTION,
    SOAP_CONTROL_BASE_URL,
    SOAP_DESCRIPTION_BASE_URL,
    SOAP_PAYLOAD,
    SOAP_RESPONSE,
    SOAP_TIMEOUT,
    SOAP_USER_AGENT,
    WS_BASE_URL,
)

_LOGGER = logging.getLogger(__name__)


class DeviceAccess:
    """Set up the device for access."""

    def __init__(self, host, jsonPort, port):
        """Initialise the utility setup."""
        self._host = host
        self._jsonPort = jsonPort
        self._port = port

    def retrieveInformation(self, rest_path, call_type=REST_GET):
        """Retrieve information from the SkyQ box."""
        try:
            if call_type == REST_GET:
                return self.http_json(rest_path)
            if call_type == REST_POST:
                return self.http_json_post(rest_path)
            if call_type == REST_DELETE:
                return self.http_json_delete(rest_path)
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ):  # as err:
            # _LOGGER.debug(f"D0010U - Connection error: {self._host} : {err}")
            return None
        # except Exception as err:
        #     _LOGGER.exception(f"X0060U - Error occurred: {self._host} : {err}")
        #     return None

    def getSoapControlURL(self, descriptionIndex):
        """Get the sky control url."""
        descriptionUrl = SOAP_DESCRIPTION_BASE_URL.format(self._host, descriptionIndex)
        headers = {"User-Agent": SOAP_USER_AGENT}
        empty_return = {"url": None, "status": "Not Found"}
        try:
            resp = requests.get(descriptionUrl, headers=headers, timeout=SOAP_TIMEOUT)
            if resp.status_code == HTTPStatus.OK:
                description = xmltodict.parse(resp.text)
                deviceType = description["root"]["device"]["deviceType"]
                if SKYCONTROL not in deviceType:
                    return empty_return

                playService = self._findPlayService(description)

                if playService is None:
                    return empty_return

                return {
                    "url": SOAP_CONTROL_BASE_URL.format(self._host, playService["controlURL"]),
                    "status": "OK",
                }
            return empty_return
        except (requests.exceptions.Timeout):
            _LOGGER.warning(f"W0020U - Control URL not accessible: {self._host} : {descriptionUrl}")
            return {"url": None, "status": "Error"}
        except (requests.exceptions.ConnectionError) as err:
            _LOGGER.exception(f"X0010U - Connection error: {self._host} : {err}")
            return empty_return
        except Exception as err:
            _LOGGER.exception(f"X0020U - Other error occurred: {self._host} : {err}")
            return empty_return

    def callSkyWebSocket(self, method):
        """Make a websocket call to the sky box."""
        try:
            ws = websocket.create_connection(WS_BASE_URL.format(self._host, method))
            response = json.loads(ws.recv())
            ws.close()
            return response
        except (TimeoutError) as err:
            _LOGGER.warning(f"W0030U - Websocket call failed: {self._host} : {method} : {err}")
            return {"url": None, "status": "Error"}
        except Exception as err:
            _LOGGER.exception(f"X0030U - Error occurred: {self._host} : {err} : {method}")
            return None

    def callSkySOAPService(self, soapControlURL, method):
        """Make a SOAP call to the sky box."""
        try:
            payload = SOAP_PAYLOAD.format(method)
            headers = {
                "Content-Type": 'text/xml; charset="utf-8"',
                "SOAPACTION": SOAP_ACTION.format(method),
            }
            resp = requests.post(
                soapControlURL,
                headers=headers,
                data=payload,
                verify=True,
                timeout=SOAP_TIMEOUT,
            )
            if resp.status_code == HTTPStatus.OK:
                xml = resp.text
                return xmltodict.parse(xml)["s:Envelope"]["s:Body"][SOAP_RESPONSE.format(method)]
            return None
        except requests.exceptions.RequestException:
            return None

    def http_json(self, path, headers=None) -> str:
        """Make an HTTP get call to the sky box."""
        response = requests.get(
            REST_BASE_URL.format(self._host, self._jsonPort, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.json()

    def http_json_post(self, path, headers=None) -> str:
        """Make an HTTP post call to the sky box."""
        response = requests.post(
            REST_BASE_URL.format(self._host, self._jsonPort, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.status_code

    def http_json_delete(self, path, headers=None) -> str:
        """Make an HTTP delete call to the sky box."""
        response = requests.delete(
            REST_BASE_URL.format(self._host, self._jsonPort, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.status_code

    def sendCommand(self, port, code):
        """Send a command to the sky box."""
        commandBytes = bytearray([4, 1, 0, 0, 0, 0, int(math.floor(224 + (code / 16))), code % 16])

        try:
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as err:
            _LOGGER.exception(f"X0040U - Failed to create socket when sending command: {self._host} : {err}")
            return

        try:
            client.connect((self._host, port))
        except Exception as err:
            _LOGGER.exception(f"X0050U - Failed to connect to client when sending command: {self._host} : {err}")
            return

        strlen = 12
        timeout = time.time() + CONNECT_TIMEOUT

        while 1:
            data = client.recv(1024)
            data = data

            if len(data) < 24:
                client.sendall(data[0:strlen])
                strlen = 1
            else:
                client.sendall(commandBytes)
                commandBytes[1] = 0
                client.sendall(commandBytes)
                client.close()
                break

            if time.time() > timeout:
                _LOGGER.error(f"E0030U - Timeout error sending command: {self._host} : {code}")
                break

    def press(self, sequence):
        """Issue the specified sequence of commands to SkyQ box."""
        if isinstance(sequence, list):
            for item in sequence:
                if item.casefold() not in COMMANDS:
                    _LOGGER.error(f"E0020 - Invalid command: {self._host} : {item}")
                    break
                self.sendCommand(self._port, COMMANDS[item.casefold()])
                time.sleep(0.5)
        elif sequence not in COMMANDS:
            _LOGGER.error(f"E0030 - Invalid command: {self._host} : {sequence}")
        else:
            self.sendCommand(self._port, COMMANDS[sequence.casefold()])

    def _findPlayService(self, description):
        services = description["root"]["device"]["serviceList"]["service"]
        if not isinstance(services, list):
            services = [services]
        playService = None
        for s in services:
            if s["serviceId"] == SKY_PLAY_URN:
                playService = s

        return playService



"""Setup skyq_remote package."""
from setuptools import setup, find_namespace_packages

from pyskyqremote.version import __version__ as version

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyskyqremote',
    version=version,
    author='Roger Selwyn',
    author_email='roger.selwyn@nomail.com',
    description='Library for Sky Q Remote',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RogerSelwyn/skyq_remote',
    license='MIT',
    packages=find_namespace_packages(exclude=['tests','manage']),
    install_requires=['requests>=2.24.0', 'websocket-client>=0.56.0', 'xmltodict>=0.12.0', 'pycountry>=20.7.3'],
    keywords='SKYQ Remote',
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.7'	
)



"""Structure of a standard EPG programme."""

import json
from dataclasses import dataclass, field
from datetime import datetime


@dataclass(order=True)
class Programme:
    """SkyQ Programme Class."""

    programmeuuid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    starttime: datetime = field(
        init=True,
        repr=True,
        compare=True,
    )
    endtime: datetime = field(
        init=True,
        repr=True,
        compare=False,
    )
    title: str = field(
        init=True,
        repr=True,
        compare=True,
    )
    season: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    episode: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    imageUrl: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelname: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    status: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    pvrid: str = "n/a"
    eid: str = "n/a"

    def __hash__(self):
        """Calculate the hash of this object."""
        return hash(self.starttime)

    def as_json(self) -> str:
        """Return a JSON string representing this Programme."""
        return json.dumps(self, cls=_ProgrammeJSONEncoder)


def ProgrammeDecoder(obj):
    """Decode programme object from json."""
    programme = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in programme and programme["__type__"] == "__programme__":
        return Programme(**programme["attributes"])
    return programme


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "starttime" in obj:
        obj["starttime"] = datetime.strptime(obj["starttime"], "%Y-%m-%dT%H:%M:%SZ")
    if "endtime" in obj:
        obj["endtime"] = datetime.strptime(obj["endtime"], "%Y-%m-%dT%H:%M:%SZ")
    return obj


class _ProgrammeJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Programme):
            attributes = {}
            for k, v in vars(obj).items():
                if isinstance(v, datetime):
                    v = v.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = v
            return {
                "__type__": "__programme__",
                "attributes": attributes,
            }



"""Methods for retrieving recording information."""

import json
import logging
from dataclasses import dataclass, field
from datetime import datetime

from ..const import (
    ALLRECORDINGS,
    RESPONSE_OK,
    REST_BOOK_PPVRECORDING,
    REST_BOOK_RECORDING,
    REST_BOOK_SERIES_RECORDING,
    REST_DELETE,
    REST_POST,
    REST_QUOTA_DETAILS,
    REST_RECORDING_DELETE,
    REST_RECORDING_DETAILS,
    REST_RECORDING_ERASE,
    REST_RECORDING_ERASE_ALL,
    REST_RECORDING_KEEP,
    REST_RECORDING_LOCK,
    REST_RECORDING_SET_LAST_PLAYED_POSITION,
    REST_RECORDING_UNDELETE,
    REST_RECORDING_UNKEEP,
    REST_RECORDING_UNLOCK,
    REST_RECORDINGS_LIST,
    REST_SERIES_LINK,
    REST_SERIES_UNLINK,
)
from .programme import Programme

_LOGGER = logging.getLogger(__name__)


class RecordingsInformation:
    """Sky Q recordings information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the recordings information class."""
        self._remoteConfig = remoteConfig

    def getRecordings(self, status, limit, offset):
        """Get the list of available Recordings."""
        recordings = set()
        resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDINGS_LIST.format(limit, offset))
        if not resp or "pvrItems" not in resp:
            _LOGGER.error(f"E0010R - Timeout retrieving recordings: {self._remoteConfig.host}")
            return Recordings(recordings)
        recData = resp["pvrItems"]
        for recording in recData:
            if recording["status"] == status or status == ALLRECORDINGS:
                built = self._buildRecording(recording)
                recordings.add(built)

        return Recordings(recordings)

    def getRecording(self, pvrId):
        """Get the recording details."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_DETAILS.format(pvrId))
        if not resp or "details" not in resp:
            _LOGGER.info(f"I0010R - Recording data not found for {pvrId}")
            return None

        recording = resp["details"]

        return self._buildRecording(recording)

    def getQuota(self):
        """Get the quota information."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_QUOTA_DETAILS)
        if not resp or "userQuotaMax" not in resp:
            _LOGGER.debug(f"D0010R - Recording data not found for {resp}")
            return None
        return Quota(resp["userQuotaMax"], resp["userQuotaUsed"])

    def bookRecording(self, eid, series):
        """Book recording for specified item."""
        resp = None
        if not series:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_BOOK_RECORDING.format(eid), REST_POST)
        else:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(
                REST_BOOK_SERIES_RECORDING.format(eid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def bookPPVRecording(self, eid, offerref):
        """Book PPV recording for specified item."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(
            REST_BOOK_PPVRECORDING.format(eid, offerref), REST_POST
        )
        if resp != RESPONSE_OK:
            return False

        return True

    def seriesLink(self, pvrid, On):
        """Series link the specified item."""
        resp = None
        if On:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_SERIES_LINK.format(pvrid), REST_POST)
        else:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_SERIES_UNLINK.format(pvrid), REST_POST)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingKeep(self, pvrid, On):
        """Keep the specified item."""
        resp = None
        if On:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_KEEP.format(pvrid), REST_POST)
        else:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_UNKEEP.format(pvrid), REST_POST)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingLock(self, pvrid, On):
        """Lock the specified item."""
        resp = None
        if On:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_LOCK.format(pvrid), REST_POST)
        else:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_UNLOCK.format(pvrid), REST_POST)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingDelete(self, pvrid, On):
        """Delete the specified item."""
        resp = None
        if On:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_DELETE.format(pvrid), REST_POST)
        else:
            resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_UNDELETE.format(pvrid), REST_POST)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingErase(self, pvrid):
        """Permanently erase the specified item."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_ERASE.format(pvrid), REST_POST)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingEraseAll(self):
        """Permanently erase the specified item."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(REST_RECORDING_ERASE_ALL, REST_DELETE)

        if resp != RESPONSE_OK:
            return False

        return True

    def recordingSetLastPlayedPosition(self, pvrid, pos):
        """Set the last played position for specified item."""
        resp = self._remoteConfig.deviceAccess.retrieveInformation(
            REST_RECORDING_SET_LAST_PLAYED_POSITION.format(pos, pvrid), REST_POST
        )
        if resp != RESPONSE_OK:
            return False

        return True

    def _buildRecording(self, recording):
        season = None
        episode = None
        starttime = None
        endtime = None
        programmeuuid = None
        channel = None
        imageUrl = None
        title = None
        status = None
        pvrid = None
        eid = None

        channel = recording["cn"]
        title = recording["t"]
        if "seasonnumber" in recording and "episodenumber" in recording:
            season = recording["seasonnumber"]
            episode = recording["episodenumber"]
        if "programmeuuid" in recording:
            programmeuuid = recording["programmeuuid"]
            imageUrl = self._remoteConfig.remoteCountry.pvr_image_url.format(str(programmeuuid))
        elif "osid" in recording:
            sid = str(recording["osid"])
            imageUrl = self._remoteCountry.buildChannelImageUrl(sid, channel)

        starttimestamp = 0
        if "ast" in recording:
            starttimestamp = recording["ast"]
        elif "st" in recording:
            starttimestamp = recording["st"]
        starttime = datetime.utcfromtimestamp(starttimestamp)

        if "finald" in recording:
            endtime = datetime.utcfromtimestamp(starttimestamp + recording["finald"])
        elif "schd" in recording:
            endtime = datetime.utcfromtimestamp(starttimestamp + recording["schd"])
        else:
            endtime = starttime

        pvrid = recording["pvrid"]
        if "oeid" in recording:
            eid = recording["oeid"]
        status = recording["status"]

        return Programme(
            programmeuuid, starttime, endtime, title, season, episode, imageUrl, channel, status, pvrid, eid
        )


@dataclass
class Recordings:
    """SkyQ Channel EPG Class."""

    programmes: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this Channel."""
        return json.dumps(self, cls=_RecordingsJSONEncoder)


def RecordingsDecoder(obj):
    """Decode channel object from json."""
    recordings = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in recordings and recordings["__type__"] == "__recordings__":
        return Recordings(programmes=recordings["programmes"], **recordings["attributes"])
    return recordings


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "starttime" in obj:
        obj["starttime"] = datetime.strptime(obj["starttime"], "%Y-%m-%dT%H:%M:%SZ")
    if "endtime" in obj:
        obj["endtime"] = datetime.strptime(obj["endtime"], "%Y-%m-%dT%H:%M:%SZ")
    if "__type__" in obj and obj["__type__"] == "__programme__":
        obj = Programme(**obj["attributes"])
    return obj


class _RecordingsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Recordings):
            type_ = "__recordings__"
            programmes = obj.programmes
            attributes = {k: v for k, v in vars(obj).items() if k not in {"programmes"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "programmes": programmes,
            }

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, Programme):
            attributes = {}
            for k, v in vars(obj).items():
                if isinstance(v, datetime):
                    v = v.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = v
            return {
                "__type__": "__programme__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, obj)  # pragma: no cover


@dataclass
class Quota:
    """SkyQ Quota Class."""

    quotaMax: int = field(
        init=True,
        repr=True,
        compare=False,
    )
    quotaUsed: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this quota info."""
        return json.dumps(self, cls=_QuotaJSONEncoder)


def QuotaDecoder(obj):
    """Decode quota object from json."""
    quota = json.loads(obj)
    if "__type__" in quota and quota["__type__"] == "__quota__":
        return Quota(**quota["attributes"])
    return quota


class _QuotaJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Quota):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__quota__",
                "attributes": attributes,
            }



"""Methods for retrieving device information."""

import json
import logging
from dataclasses import dataclass, field

from ..const import KNOWN_COUNTRIES, REST_PATH_DEVICEINFO, REST_PATH_SYSTEMINFO, UPNP_GET_TRANSPORT_INFO

_LOGGER = logging.getLogger(__name__)


class DeviceInformation:
    """Sky Q device information retrieval methods."""

    def __init__(self, remoteConfig):
        """Initialise the device information class."""
        self._remoteConfig = remoteConfig
        self._deviceAccess = remoteConfig.deviceAccess
        self._port = remoteConfig.port

    def getTransportInformation(self):
        """Get the transport information from the SkyQ box."""
        return self._deviceAccess.callSkySOAPService(self._remoteConfig.soapControlURL, UPNP_GET_TRANSPORT_INFO)

    def getSystemInformation(self):
        """Get the system information from the SkyQ box."""
        return self._deviceAccess.retrieveInformation(REST_PATH_SYSTEMINFO)

    def getDeviceInformation(self, overrideCountry):
        """Get the device information from the SkyQ box."""
        deviceInfo = self._deviceAccess.retrieveInformation(REST_PATH_DEVICEINFO)
        if not deviceInfo:
            return None

        systemInfo = self.getSystemInformation()
        ASVersion = deviceInfo["ASVersion"]
        IPAddress = deviceInfo["IPAddress"]
        countryCode = deviceInfo["countryCode"]
        gateway = deviceInfo["gateway"]
        hardwareModel = systemInfo["hardwareModel"]
        deviceType = systemInfo["deviceType"]
        hardwareName = deviceInfo["hardwareName"]
        manufacturer = systemInfo["manufacturer"]
        modelNumber = deviceInfo["modelNumber"]
        serialNumber = deviceInfo["serialNumber"]
        versionNumber = deviceInfo["versionNumber"]
        bouquet = deviceInfo["bouquet"]
        subbouquet = deviceInfo["subbouquet"]

        epgCountryCode = overrideCountry or countryCode.upper()
        if not epgCountryCode:
            _LOGGER.error(f"E0010 - No country identified: {self._host}")
            return None

        if epgCountryCode in KNOWN_COUNTRIES:
            epgCountryCode = KNOWN_COUNTRIES[epgCountryCode]

        return Device(
            ASVersion,
            IPAddress,
            countryCode,
            epgCountryCode,
            hardwareModel,
            hardwareName,
            deviceType,
            gateway,
            manufacturer,
            modelNumber,
            serialNumber,
            versionNumber,
            bouquet,
            subbouquet,
        )

    def getSoapControlURL(self):
        """Get the soapcontrourl for the SkyQ box."""
        url_index = 0
        soapControlURL = None
        while soapControlURL is None and url_index < 50:
            soapControlURL = self._deviceAccess.getSoapControlURL(url_index)["url"]
            url_index += 1

        return soapControlURL


@dataclass
class Device:
    """SkyQ Device Class."""

    ASVersion: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    IPAddress: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    countryCode: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    epgCountryCode: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    hardwareModel: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    hardwareName: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    deviceType: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    gateway: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    manufacturer: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    modelNumber: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    serialNumber: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    versionNumber: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    bouquet: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    subbouquet: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this device info."""
        return json.dumps(self, cls=_DeviceJSONEncoder)


def DeviceDecoder(obj):
    """Decode programme object from json."""
    device = json.loads(obj)
    if "__type__" in device and device["__type__"] == "__device__":
        return Device(**device["attributes"])
    return device


class _DeviceJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Device):
            attributes = {k: v for k, v in vars(obj).items()}
            return {
                "__type__": "__device__",
                "attributes": attributes,
            }



"""Constants for pyskyqremote- DE."""

SCHEDULE_URL = "https://www.sky.de/sgtvg/service/getBroadcasts"
LIVE_IMAGE_URL = "https://www.sky.de{0}"
PVR_IMAGE_URL = "https://www.sky.de{0}"
CHANNEL_IMAGE_URL = "https://www.sky.de{0}"
TIMEZONE = "Europe/Berlin"
CHANNEL_URL = "https://raw.githubusercontent.com/RogerSelwyn/skyq_remote/master/pyskyqremote/country/channels-de.json"



"""DE specific code."""
import json
import logging
from datetime import datetime, timedelta, timezone

import pytz
import requests

from ..classes.programme import Programme
from ..const import RESPONSE_OK, SKY_STATUS_LIVE
from .const_de import CHANNEL_IMAGE_URL, CHANNEL_URL, LIVE_IMAGE_URL, PVR_IMAGE_URL, SCHEDULE_URL, TIMEZONE

_LOGGER = logging.getLogger(__name__)


class SkyQCountry:
    """DE specific SkyQ."""

    def __init__(self):
        """Initialise DE remote."""
        self.pvr_image_url = PVR_IMAGE_URL
        self._channellist = None

        self._getChannels()

    def getEpgData(self, sid, channelno, channelName, epgDate):
        """Get EPG data for DE."""
        return self._getData(sid, channelno, channelName, epgDate)

    def buildChannelImageUrl(self, sid, channelname):
        """Build the channel image URL."""
        for channel in self._channellist:
            if str(channel["sid"]) == str(sid):
                return CHANNEL_IMAGE_URL.format(channel["clu"])

    def _getData(self, sid, channelno, channelName, epgDate):
        cid = None
        for channel in self._channellist:
            if str(channel["sid"]) == str(sid):
                cid = channel["ci"]

        milliDate = int(epgDate.replace(tzinfo=timezone.utc).timestamp() * 1000)
        berlinDT = epgDate.replace(tzinfo=pytz.utc).astimezone(pytz.timezone(TIMEZONE))
        berlinDate = berlinDT.strftime("%Y-%m-%dT")

        epgUrl = SCHEDULE_URL
        programmes = set()

        headers = {
            "Content-Type": 'application/json; charset="utf-8"',
        }
        payload = json.dumps(
            {
                "d": milliDate,
                "lt": 6,
                "t": 0,
                "s": 0,
                "pn": 1,
                "sto": 10,
                "epp": 50,
                "cil": [cid],
            }
        )

        resp = requests.post(
            epgUrl,
            headers=headers,
            data=payload,
            verify=True,
            timeout=10,
        )
        epgData = resp.json()["el"] if resp.status_code == RESPONSE_OK else None
        if epgData is None:
            return programmes

        if len(epgData) == 0:
            return programmes

        for p in epgData:
            starttimede = datetime.strptime(berlinDate + p["bst"], "%Y-%m-%dT%H:%M")
            starttime = starttimede.replace(tzinfo=berlinDT.tzinfo).astimezone(pytz.utc).replace(tzinfo=None)
            endtime = starttime + timedelta(minutes=p["len"])
            title = p["et"]
            season = None
            # if "seasonnumber" in p:
            #     if p["seasonnumber"] > 0:
            #         season = p["seasonnumber"]
            episode = None
            # if "episodenumber" in p:
            #     if p["episodenumber"] > 0:
            #         episode = p["episodenumber"]
            programmeuuid = None
            imageUrl = None
            # if "programmeuuid" in p:
            programmeuuid = str(p["ei"])
            if "pu" in p:
                imageUrl = LIVE_IMAGE_URL.format(p["pu"])
            elif "clu" in p:
                imageUrl = LIVE_IMAGE_URL.format(p["clu"])

            programme = Programme(
                programmeuuid,
                starttime,
                endtime,
                title,
                season,
                episode,
                imageUrl,
                channelName,
                SKY_STATUS_LIVE,
            )
            programmes.add(programme)

        return programmes

    def _getChannels(self):
        resp = requests.get(CHANNEL_URL)
        if resp.status_code == RESPONSE_OK:
            self._channellist = resp.json()
